from __future__ import print_function, division, absolute_import
import time
from events import Events
from websocket import ABNF
import websocket
from hkube_python_wrapper.util.encoding import Encoding
from threading import Thread


class WebsocketClient(Thread):
    def __init__(self, msg_queue, encoding, url):
        Thread.__init__(self, name='WebsocketClient')
        self.daemon = True
        self.events = Events()
        self._msg_queue = msg_queue
        self._ws = None
        self._reconnectInterval = 0.1
        self._active = True
        self._firstConnect = False
        self._encoding = Encoding(encoding)
        self._ws_opcode = ABNF.OPCODE_BINARY if self._encoding.isBinary else ABNF.OPCODE_TEXT
        self._url = url
        print('Initialized socket with {encoding} encoding'.format(encoding=encoding))

    def on_message(self, message):
        decoded = self._encoding.decode(message, plainEncode=True)
        command = decoded["command"]
        data = decoded.get("data", None)
        print('got message from worker: {command}'.format(command=command))
        self._msg_queue.put((command, data))

    def on_error(self, error):
        if self._firstConnect:
            print(error)

    def on_close(self):
        self.events.on_disconnect()

    def on_open(self):
        self._firstConnect = True
        self.events.on_connection()

    def send(self, message):
        print('sending message to worker: {command}'.format(**message))
        self._ws.send(self._encoding.encode(message, plainEncode=True), opcode=self._ws_opcode)

    def run(self):
        self._startWS(self._url)

    def _startWS(self, url):
        self._ws = websocket.WebSocketApp(
            url,
            on_message=self.on_message,
            on_error=self.on_error,
            on_open=self.on_open,
            on_close=self.on_close)
        while self._active:
            try:
                self._ws.run_forever()
                time.sleep(self._reconnectInterval)
            except Exception:
                pass

    def shutDown(self):
        self._active = False
        if self._ws:
            self._ws.close()



import zmq
import time
from .consts import consts

context = zmq.Context()


class ZMQRequest(object):
    def __init__(self, reqDetails):
        self.socket = self._createSocket(host=reqDetails['host'], port=reqDetails['port'])
        self.ping_socket = self._createSocket(host=reqDetails['host'], port=int(reqDetails['port'])+1)
        self.content = reqDetails['content']
        self.timeout = int(reqDetails['timeout'])
        self.networkTimeout = int(reqDetails['networkTimeout'])
        self.pingTime = 1e10

    def invokeAdapter(self):
        print('sending ping')
        pingStart = time.time()
        self.ping_socket.send(consts.zmq.ping)
        result = self.ping_socket.poll(self.networkTimeout)
        if (result):
            there = self.ping_socket.recv()
            if (there == consts.zmq.pong):
                print('got pong')
                pingEnd = time.time()
                self.pingTime = (pingEnd-pingStart)*1000
                self.socket.send(self.content)
                result = self.socket.poll(self.timeout)
                if (result):
                    message = self.socket.recv_multipart()
                    return message
                print('Timed out')
                raise Exception('Timed out:' + str(self.timeout))
        print('Ping timed out ({timeout}) to {conn}'.format(timeout=self.networkTimeout, conn=self.connStr))
        raise Exception('Ping timed out ({timeout}) to {conn}'.format(timeout=self.networkTimeout, conn=self.connStr))

    def close(self):
        self.socket.close()

    def _createSocket(self, host, port):
        connStr = 'tcp://' + host + ':' + str(port)
        print('creating socket to: {s}'.format(s=connStr))
        socket = context.socket(zmq.REQ)
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(connStr)
        return socket



__title__ = 'hkube_python_wrapper'
__description__ = 'Hkube Python Wrapper'
__url__ = 'https://github.com/kube-HPC/python-wrapper.hkube'
__author__ = 'Hkube'
__author_email__ = 'hkube.dev@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Hkube'



from __future__ import print_function, division, absolute_import
from hkube_python_wrapper.storage.task_output_manager import TaskOutputManager
from hkube_python_wrapper.storage.base_storage_manager import BaseStorageManager
from hkube_python_wrapper.storage.fs_adapter import FSAdapter
from hkube_python_wrapper.storage.s3_adapter import S3Adapter

adapterTypes = {
    'fs': FSAdapter,
    's3': S3Adapter
}


class StorageManager():
    def __init__(self, config):
        storageType = config["type"]
        encoding = config["encoding"]
        adapterConfig = config[storageType]
        adapterType = adapterTypes.get(storageType)
        adapter = adapterType(adapterConfig)
        self.hkube = TaskOutputManager(adapter, config)
        self.storage = BaseStorageManager(adapter)
        print('init {type} storage client with {encoding} encoding'.format(type=storageType, encoding=encoding))



import os
import random
from hkube_python_wrapper.util.encoding import Encoding


def test_json_encoding():
    encoding = Encoding('json')
    sizes = [100, 200, 300]
    for size in sizes:
        data = createObjectJson(size)
        encoded = encoding.encode(data)
        decoded = encoding.decode(encoded)
        assert data == decoded


def test_bson_encoding():
    encoding = Encoding('bson')
    sizes = [100, 200, 300]
    for size in sizes:
        data = createObject(size * 1000000, size)
        encoded = encoding.encode(data)
        decoded = encoding.decode(encoded)
        assert data == decoded


mb = 1024 * 1024


def create_bytearray(sizeBytes):
    return b'\xdd' * (sizeBytes)


def test_msgpack_encoding_bytearray():
    encoding = Encoding('msgpack')
    data = create_bytearray(20)
    encoded = encoding.encode(data)
    decoded = encoding.decode(encoded)
    assert data == decoded


def test_msgpack_encoding_bytearray_separatelly():
    encoding = Encoding('msgpack')
    data = create_bytearray(20)
    header, encoded = encoding.encode_separately(data)
    decoded = encoding.decode_separately(header, encoded)
    assert data == decoded


def xtest_msgpack_encoding_string():
    encoding = Encoding('msgpack')
    data = create_bytearray(20)
    encoded = encoding.encode(data)
    decoded = encoding.decode(data)
    assert data2 == decoded


def test_msgpack_encoding():
    encoding = Encoding('msgpack')
    sizes = [1, 2, 3]
    for size in sizes:
        data = createObject(size * 1000000, size)
        encoded = encoding.encode(data)
        decoded = encoding.decode(encoded)
        assert data == decoded


def test_msgpack_encodingx():
    use_bin_type_false = 'use_bin_type'
    encoding = Encoding('msgpack')
    data = {
        'taskId': 'taskid',
        'prop': 'blasds'
    }
    # encoded = encoding.encode(data)

    # with open(use_bin_type_false, 'wb') as f:
    #     f.write(encoded)

    # with open(use_bin_type_false, 'rb') as f:
    #     result = f.read()

    # decoded = encoding.decode(result)
    # assert decoded == decoded


def randomString(n):
    min_lc = ord(b'a')
    len_lc = 26
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = min_lc + b % len_lc  # convert 0..255 to 97..122
    return ba.decode("utf-8")


def randomInt(size):
    return random.sample(range(0, size), size)


def createObject(sizeBytes, sizeRandom):
    obj = {
        "bytesData": bytearray(b'\xdd' * (sizeBytes)),
        "anotherBytesData": bytearray(sizeBytes),
        "randomString": randomString(sizeRandom),
        "randomIntArray": randomInt(sizeRandom),
        "dataString": randomString(sizeRandom),
        "bool": False,
        "anotherBool": False,
        "nestedObj": {
            "dataString": randomString(sizeRandom),
            "randomIntArray": randomInt(sizeRandom)
        }
    }
    return obj


def createObjectJson(sizeRandom):
    obj = {
        "randomString": randomString(sizeRandom),
        "randomIntArray": randomInt(sizeRandom),
        "dataString": randomString(sizeRandom),
        "bool": False,
        "anotherBool": False,
        "nestedObj": {
            "dataString": randomString(sizeRandom),
            "randomIntArray": randomInt(sizeRandom)
        }
    }
    return obj



from websocket_server import WebsocketServer
from hkube_python_wrapper.util.encoding import Encoding
from tests.mocks import mockdata
from tests.configs import config


class WebSocketServerClass:
    def __init__(self, encoding, server):
        self._server = server
        self._server.set_fn_new_client(self.handleConnected)
        self._server.set_fn_client_left(self.handleDisconnected)
        self._server.set_fn_message_received(self.handleMessage)
        self._encoding = Encoding(encoding)
        self._commands = {
            "initialized":  {
                'command': "start",
                'data': lambda x: x
            },
            "startAlgorithmExecution": {
                'command': "algorithmExecutionDone",
                'data': lambda x: {
                    'execId': x.get('execId'),
                    'response': x.get('input')
                }
            },
            "startStoredSubPipeline": {
                'command': "subPipelineDone",
                'data': lambda x: {
                    'subPipelineId': x.get('subPipelineId'),
                    'response': x.get('subPipeline').get('flowInput')
                }
            }
        }

    def handleMessage(self, client, server, message):
        decoded = self._encoding.decode(message, plainEncode=True)
        command = decoded["command"]
        data = decoded.get("data", None)
        commandBack = self._commands.get(command)
        if(commandBack):
            msgBack = {
                "command": commandBack["command"],
                "data": commandBack["data"](data)
            }
            self.sendMsgToClient(client, msgBack)

    def handleConnected(self, client, server):
        # print('ws connected')
        self.sendMsgToClient(client, {'command': 'initialize', 'data': mockdata.initData})

    def handleDisconnected(self, client, server):
        # print('ws disconnected')
        pass

    def sendMsgToClient(self, client, data):
        self._server.send_message(client, self._encoding.encode(data, plainEncode=True))


def startWebSocketServer(options):
    port = options["port"]
    encoding = options["encoding"]
    server = WebsocketServer(int(port))
    wss = WebSocketServerClass(encoding, server)
    server.run_forever()


if __name__ == "__main__":
    startWebSocketServer(config.socket)




import os
import sys
import subprocess

ws_server = None


def pytest_report_collectionfinish(config, startdir, items):
    global ws_server
    python_path = ":".join(sys.path)[1:]
    environ = {'PYTHONPATH': python_path}
    copyEnv = os.environ.copy()
    copyEnv.update(environ)
    ws_server = subprocess.Popen(
        ['python', 'tests/mock_ws_server.py'], env=copyEnv)


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    # print('pytest_configure')


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    # print('pytest_sessionstart', session)

    python_path = ":".join(sys.path)[1:]
    environ = {'PYTHONPATH': python_path}
    copyEnv = os.environ.copy()
    copyEnv.update(environ)
    # print('python_path: ' + python_path)


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """
    global ws_server
    if ws_server is not None:
        print('killing ws_server')
        ws_server.kill()
    ws_server = None
    # print('pytest_sessionfinish')


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    global ws_server
    if ws_server is not None:
        ws_server.kill()
    ws_server = None
    # print('pytest_unconfigure')



import pytest
from hkube_python_wrapper.tracing import Tracer
from tests.configs import config
import logging
import time
import requests
import random
import string

def randomString(stringLength=8):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


def get_jaeger_trace(service_name='algorithm', operation='',limit=1):
    params={
        'service':service_name,
        'operation':operation,
        'limit':limit
        }
    r = requests.get('http://localhost:16686/api/traces',params)
    return r.json()


log_level = logging.DEBUG
logging.getLogger('').handlers = []
logging.basicConfig(format='%(asctime)s %(message)s', level=log_level)

def test_tracer_fail_init():
    with pytest.raises(TypeError):
        tracer = Tracer()

def test_tracer_init_empty_config():
    with pytest.raises(AttributeError):
        tracer = Tracer(None)

def test_tracer_init_config():
    tracer = Tracer(config.tracer)
    assert tracer.tracer is not None
    assert tracer.tracer.service_name == config.tracer.get("service_name")
    tracer.close()

def test_tracer_start_span():
    tracer = Tracer(config.tracer)
    op1=randomString()
    with tracer.tracer.start_span(operation_name=op1) as span1:
        assert span1 is not None
    time.sleep(2)
    tracer.close()
    r=get_jaeger_trace(operation=op1)
    assert r['data'][0]
        
def test_tracer_start_child_span():
    tracer = Tracer(config.tracer)
    op1=randomString()
    op2=randomString()
    with tracer.tracer.start_active_span(operation_name=op1) as span1:
        with tracer.tracer.start_active_span(operation_name=op2) as span2:
            assert span1.span is not None
            assert span2.span is not None
            assert span2.span.parent_id is span1.span.span_id
    time.sleep(2)
    tracer.close()
    r=get_jaeger_trace(operation=op1)
    assert r['data'][0]
    op1Trace=r['data'][0]['traceID']
    r=get_jaeger_trace(operation=op2)
    assert r['data'][0]
    op2Trace=r['data'][0]['traceID']
    assert op1Trace==op2Trace








taskId1 = 'task_1'
taskId2 = 'task_2'

dataTask1 = {
    'level1': {
        'level2': {
            'value1': 'd2_l1_l2_value_1',
            'value2': 'd2_l1_l2_value_2',
        },
        'value1': 'd2_l1_value_1'
    },
    'value1': 'd2_value_1'
}

dataTask2 = bytearray(b'\xdd'*(100))
dataTask2_original = bytearray(b'\xdd'*(100))

initData = {
    'jobId': 'jobId',
    'taskId': 'taskId',
    'input': [1, False, None],
    'nodeName': 'green',
    'spanId': {"uber-trace-id": "a0aa0bab5bfde7a6:7e187ec65fb04e0e:37bc77758a09b6a4:1"}
}





import math
import functools


def percentile(N, percent, key=lambda x: x):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values. Note N MUST BE already sorted.
    @parameter percent - a float value from 0 to 100
    @parameter key - optional key function to compute value from each element of N.

    @return - the percentile of the values
    """
    if not N:
        return None
    percent = percent/100.0
    k = (len(N)-1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return key(N[int(k)])
    d0 = key(N[int(f)]) * (c-k)
    d1 = key(N[int(c)]) * (k-f)
    return d0+d1


# median is 50th percentile.
median = functools.partial(percentile, percent=0.5)


def print_percentiles(values, percentiles=[10, 25, 50, 75, 90, 95], title=None):
    if not values:
        return
    if (title):
        print(title)
        print('-'*len(title))
    for p in percentiles:
        print(str(p) + ' - ' + str(percentile(values, p)))



from hkube_python_wrapper.communication.zmq.ZMQServers import ZMQServers
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.util.decorators import timing
from hkube_python_wrapper.cache.caching import Cache


class DataServer:

    def __init__(self, config):
        self._cache = Cache(config)
        self._host = config['host']
        self._port = config['port']
        self._encodingType = config['encoding']
        self._encoding = Encoding(self._encodingType)

        self._adapter = ZMQServers(self._port, self._createReply, config.get('num_threads', 5))
        self.notAvailable = self._encoding.encode_separately(
            self._createError('notAvailable', 'taskId notAvailable'))

    def listen(self):
        print('discovery serving on {host}:{port} with {encoding} encoding'.format(
            host=self._host, port=self._port, encoding=self._encodingType))
        self._adapter.listen()

    @timing
    def _createReply(self, message):
        try:
            decoded = self._encoding.decode(message, plainEncode=True)
            tasks = decoded.get('tasks')
            resultsAsTuple = self.getDataByTaskId(tasks)

        except Exception as e:
            result = self._createError('unknown', str(e))
            header, encoded = self._encoding.encode_separately(result)
            return [header, encoded]
        parts = []
        for header, content in resultsAsTuple:
            parts.append(header)
            parts.append(content)
        return parts

    def getDataByTaskId(self, tasks):
        results = []
        for task in tasks:
            if (task not in self._cache):
                result = self.notAvailable
            else:
                result = self._cache.getWithHeader(task)
            results.append(result)
        return results

    def setSendingState(self, taskId, header, encoded, size):
        return self._cache.update(taskId, encoded, size=size, header=header)

    def _createError(self, code, message):
        return {'hkube_error': {'code': code, 'message': message}}

    def isLocal(self, host, port):
        return host == self._host and port == self._port

    def isServing(self):
        return self._adapter.isServing()

    def shutDown(self):
        self._adapter.close()



from threading import Thread
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.wrapper.data_adapter import DataAdapter
from hkube_python_wrapper.communication.DataServer import DataServer
from tests.configs import config

dataAdapter = DataAdapter(config)

jobId = 'jobId-328901800'
taskId1 = 'taskId-328901801'
taskId2 = 'taskId-328901802'
taskId3 = 'taskId-328901803'
taskId4 = 'taskId-328901804'

array = [42, 37, 89, 95, 12, 126, 147]
obj1 = {'data1': {'array1': array}}
obj2 = {'data2': {'array2': array}}
obj3 = {'data3': {'array3': array}}
obj4 = {'data4': {'array4': array}}

data1 = dataAdapter.encode(obj1)
data2 = dataAdapter.encode(obj2)
data3 = dataAdapter.encode(obj3)
data4 = dataAdapter.encode(obj4)

storageInfo1 = dataAdapter.setData({'jobId': jobId, 'taskId': taskId1, 'data': data1})
storageInfo2 = dataAdapter.setData({'jobId': jobId, 'taskId': taskId2, 'data': data2})
storageInfo3 = dataAdapter.setData({'jobId': jobId, 'taskId': taskId3, 'data': data3})
storageInfo4 = dataAdapter.setData({'jobId': jobId, 'taskId': taskId4, 'data': data4})

inputArgs = [
    {'data': '$$guid-1'},
    {'prop': ['$$guid-2']},
    [{'prop': '$$guid-3'}],
    ['$$guid-4'],
    '$$guid-5',
    'test-param',
    True,
    None,
    12345
]

flatInput = {
    '0.data': '$$guid-1',
    '1.prop.0': '$$guid-2',
    '2.0.prop': '$$guid-3',
    '3.0': '$$guid-4',
    '4': '$$guid-5',
    '5': 'test-param',
    '6': True,
    '7': None,
    '8': 12345
}

storage = {
    'guid-1': {'storageInfo': storageInfo1, 'path': 'data1'},
    'guid-2': {'storageInfo': storageInfo2, 'path': 'data2.array2'},
    'guid-3': {'storageInfo': storageInfo3, 'path': 'data3.array3.4'},
    'guid-4': {'storageInfo': storageInfo4},
    'guid-5': {'storageInfo': storageInfo4, 'path': 'data4.array4'}
}

discovery = dict(config.discovery)
discovery.update({"port": 9025})
ds = DataServer(discovery)
ds.listen()
ds.setSendingState(taskId1, None, obj1, 10)
ds.setSendingState(taskId2, None, obj2, 10)
ds.setSendingState(taskId3, None, obj3, 10)
ds.setSendingState(taskId4, None, obj4, 10)


def test_get_data_no_storage():
    result = dataAdapter.getData({'input': inputArgs})
    assert result == inputArgs


def test_get_data_no_input():
    result = dataAdapter.getData({'storage': storage})
    assert result is None


def test_get_data():
    result = dataAdapter.getData({'input': inputArgs, 'flatInput': flatInput, 'storage': storage})
    assert result[0]['data']['array1'] == array
    assert result[1]['prop'][0] == array
    assert result[2][0]['prop'] == array[4]
    assert result[3][0] == obj4
    assert result[4] == array
    assert result[5] == inputArgs[5]
    assert result[6] == inputArgs[6]
    assert result[7] == inputArgs[7]
    assert result[8] == inputArgs[8]


def xtest_get_batch_request_success():
    inputArgs = [
        '$$guid-1',
    ]
    flatInput = {
        '0': '$$guid-5'
    }
    storage = {'guid-5': [{'discovery': discovery, 'tasks': [taskId2, taskId3, taskId4]},
                          {'discovery': discovery, 'tasks': [taskId2, taskId3, taskId4]},
                          {'discovery': discovery, 'tasks': [taskId2, taskId3, taskId4]},
                          {'discovery': discovery, 'tasks': [taskId2, taskId3, taskId4]}]}

    result = dataAdapter.getData({'input': inputArgs, 'flatInput': flatInput, 'storage': storage})
    assert result[0] == [obj2, obj3, obj4, obj2, obj3, obj4, obj2, obj3, obj4, obj2, obj3, obj4]


def test_get_batch_request_with_errors():
    inputArgs = [
        '$$guid-1',
    ]
    flatInput = {
        '0': '$$guid-5'
    }
    storage = {
        'guid-5': [{'discovery': discovery, 'tasks': [taskId1, taskId2, taskId3, taskId4]}]
    }

    result = dataAdapter.getData({'jobId': jobId, 'input': inputArgs, 'flatInput': flatInput, 'storage': storage})
    assert result[0] == [obj1, obj2, obj3, obj4]


def test_get_batch_request_with_storage_fallback():
    ds.shutDown()
    inputArgs = [
        '$$guid-1',
    ]
    flatInput = {
        '0': '$$guid-5'
    }
    storage = {
        'guid-5': [{'discovery': discovery, 'tasks': [taskId1, taskId2, taskId3, taskId4]}]
    }

    result = dataAdapter.getData({'jobId': jobId, 'input': inputArgs, 'flatInput': flatInput, 'storage': storage})
    assert result[0] == [obj1, obj2, obj3, obj4]


def test_set_data():
    result = dataAdapter.setData({'jobId': jobId, 'taskId': taskId1, 'data': data1})
    assert result['path'].find(jobId) != -1


def test_createMetadata_no_path():
    array = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    nodeName = 'green'
    savePaths = [nodeName + '.no.such']
    result = dataAdapter.createMetadata({
        'nodeName': nodeName,
        'data': {"prop": array},
        'savePaths': savePaths
    })
    assert typeCheck.isDict(result)


def test_createMetadata_array():
    array = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    nodeName = 'green'
    savePaths = [nodeName + '.prop.5']
    result = dataAdapter.createMetadata({
        'nodeName': nodeName,
        'data': {"prop": array},
        'savePaths': savePaths
    })
    assert result["green.prop.5"] == {"type": "int"}


def test_createMetadata_bytes():
    sizeBytes = 100 * 1000000
    bytesData = bytearray(sizeBytes)
    nodeName = 'green'
    savePaths = [nodeName + '.prop']
    result = dataAdapter.createMetadata({
        'nodeName': nodeName,
        'data': {"prop": bytesData},
        'savePaths': savePaths
    })
    assert result["green.prop"] == {"type": "bytearray"}


def test_createStorageInfo():
    savePaths = ['green.data1.array1']
    result = dataAdapter.createStorageInfo({
        'jobId': jobId,
        'taskId': taskId1,
        'nodeName': 'green',
        'data': obj1,
        'savePaths': savePaths
    })
    metadata = result['metadata'][savePaths[0]]

    assert metadata.get('size') == len(array)
    assert metadata.get('type') == 'array'
    assert result['storageInfo']['path'].find(jobId) != -1



import os
import sys
import errno
PY3 = sys.version_info[0] == 3


class FSAdapter:
    def __init__(self, config):
        self.basePath = config['baseDirectory']

    def init(self, options):
        path = options["path"]
        return self.ensure_dir(path)

    def put(self, options):
        filePath = self.getPath(self.basePath, options['path'])
        self.ensure_dir(filePath)
        with open(filePath, 'wb') as f:
            f.write(options['data'])
        return {'path': options['path']}

    def get(self, options):
        filePath = self.getPath(self.basePath, options['path'])
        result = None
        with open(filePath, 'rb') as f:
            result = f.read()
        return result

    def list(self, options):
        filePath = self.basePath + os.path.sep + options['path']
        if not (os.path.exists(filePath)):
            return None
        recursive_files_in_dir = []
        for r, d, f in os.walk(filePath):  # pylint: disable=unused-variable
            files_in_dir = []
            relativePath = r.replace(self.basePath, '')
            for fname in f:
                files_in_dir.append({'path': relativePath + os.path.sep + fname})
            recursive_files_in_dir = recursive_files_in_dir + files_in_dir
        return recursive_files_in_dir

    def delete(self, options):
        filePath = self.getPath(self.basePath, options['path'])
        os.remove(filePath)

    def listPrefix(self, options):
        filePath = self.basePath + os.path.sep + options['path']
        if not (os.path.exists(filePath)):
            return None
        for r, d, f in os.walk(filePath):  # pylint: disable=unused-variable
            files_in_dir = []
            relativePath = r.replace(self.basePath, '')
            for fname in f:
                files_in_dir.append(relativePath + os.path.sep + fname)
            break
        return files_in_dir

    @staticmethod
    def ensure_dir(dirName):
        return FSAdapter.ensure_dir_py3(dirName) if PY3 else FSAdapter.ensure_dir_py27(dirName)

    @staticmethod
    def ensure_dir_py3(dirName):
        d = os.path.dirname(dirName)
        # pylint: disable=unexpected-keyword-arg
        os.makedirs(d, exist_ok=True)
        return os.path.exists(dirName)

    @staticmethod
    def ensure_dir_py27(dirName):
        d = os.path.dirname(dirName)
        try:
            os.makedirs(d)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return os.path.exists(dirName)

    @staticmethod
    def getPath(base, dirName):
        return base + os.path.sep + dirName



from __future__ import print_function, division, absolute_import
import multiprocessing
import concurrent.futures
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.util.decorators import timing, trace
from hkube_python_wrapper.util.object_path import getPath, setPath
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.storage.storage_manager import StorageManager
from hkube_python_wrapper.communication.DataRequest import DataRequest
from hkube_python_wrapper.cache.caching import Cache
from ..config import config
from hkube_python_wrapper.util.percentile import print_percentiles


class DataAdapter:
    def __init__(self, options, dataServer=None):
        self._dataServer = dataServer
        self._storageCache = Cache(config.storage)
        self._encoding = Encoding(options.storage['encoding'])
        self._storageManager = StorageManager(options.storage)
        self._requestEncoding = options.storage['encoding']
        self._requestTimeout = options.discovery['timeout']
        self._networkTimeout = options.discovery['networkTimeout']
        self._maxWorkers = min(32, (multiprocessing.cpu_count() or 1) + 4)
        print('using {workers} workers for DataAdapter'.format(workers=self._maxWorkers))

    def encode(self, value):
        return self._encoding.encode(value)

    def encode_separately(self, value):
        return self._encoding.encode_separately(value)

    def decode(self, value):
        return self._encoding.decode(value)

    @trace()
    def getData(self, options):
        jobId = options.get('jobId')
        inputArgs = options.get('input')
        flatInput = options.get('flatInput')
        storage = options.get('storage')

        if (not flatInput):
            return inputArgs

        for k, v in flatInput.items():
            if self._isStorage(v):
                key = v[2:]
                link = storage.get(key, None)
                if (link is None):
                    raise Exception('unable to find storage key')

                if (typeCheck.isList(link)):
                    data = self.batchRequest(link, jobId)
                else:
                    data = self.tryGetDataFromPeerOrStorage(link)

                setPath(inputArgs, k, data)
        try:
            print_percentiles(DataRequest.ping_times, title='Ping Times')
        except Exception:
            pass

        DataRequest.ping_times = []

        return inputArgs

    def _isStorage(self, value):
        return typeCheck.isString(value) and value.startswith('$$')

    @trace()
    def setData(self, options):
        jobId = options.get('jobId')
        taskId = options.get('taskId')
        data = options.get('data')
        result = self._storageManager.hkube.put(jobId, taskId, data)
        return result

    @timing
    def batchRequest(self, options, jobId):
        batchResponse = []
        for d in options:
            d.update({"jobId": jobId})

        with concurrent.futures.ThreadPoolExecutor(max_workers=self._maxWorkers) as executor:
            for out in executor.map(self._batchRequest, options):
                batchResponse += out

        return batchResponse

    def _batchRequest(self, options):
        batchResponse = []
        jobId = options.get('jobId')
        tasks = options.get('tasks')
        dataPath = options.get('path')
        storageInfo = options.get('storageInfo')
        if (storageInfo):
            storageResult = self._getFromCacheOrStorage(storageInfo, dataPath, storageInfo.get("path"))
            batchResponse.append(storageResult)
            return batchResponse
        tasksNotInCache, batchResponse = self._storageCache.getAll(tasks)
        if (tasksNotInCache):
            options['tasks'] = tasksNotInCache
            results = self._getFromPeer(options)
            for i, item in enumerate(results):
                size, content = item
                peerError = self._getPeerError(content)
                taskId = tasksNotInCache[i]
                if (peerError):
                    storageData = self._getDataForTask(jobId, taskId, dataPath)
                    batchResponse.append(storageData)
                else:
                    self._storageCache.update(taskId, content, size)
                    content = self._getPath(content, dataPath)
                    batchResponse.append(content)

        return batchResponse

    def _getDataForTask(self, jobId, taskId, dataPath):
        path = self._storageManager.hkube.createPath(jobId, taskId)
        return self._getFromCacheOrStorage({'path': path}, dataPath, taskId)

    def tryGetDataFromPeerOrStorage(self, options):
        dataPath = options.get('path')
        storageInfo = options.get('storageInfo')
        discovery = options.get('discovery')
        data = None
        hasResponse = False
        if (discovery):
            cacheId = options.get('taskId')
        else:
            cacheId = storageInfo.get('path')
        data = self._getFromCache(cacheId, dataPath)
        if not (data):
            if (discovery):
                size, data = self._getFromPeer(options)[0]
                peerError = self._getPeerError(data)
                hasResponse = not peerError
                data = None if peerError else data
                if (hasResponse):
                    self._setToCache(cacheId, data, size)
                    data = self._getPath(data, dataPath)
            if (not hasResponse and storageInfo):
                data = self._getFromCacheOrStorage(storageInfo, dataPath, cacheId)

        return data

    @trace(name='getFromPeer')
    @timing
    def _getFromPeer(self, options):
        taskId = options.get('taskId')
        tasks = [taskId] if taskId else options.get('tasks')
        discovery = options.get('discovery')
        port = discovery.get('port')
        host = discovery.get('host')

        if (self._dataServer and self._dataServer.isLocal(host, port)):
            dataList = self._dataServer.getDataByTaskId(tasks)
            responses = []
            for data in dataList:
                responses.append((len(data), self.decode(data)))
        else:
            request = {
                'address': {
                    'port': port,
                    'host': host
                },
                'tasks': tasks,
                'encoding': self._requestEncoding,
                'timeout': self._requestTimeout,
                'networkTimeout': self._networkTimeout
            }
            dataRequest = DataRequest(request)
            responses = dataRequest.invoke()
        return responses

    def _getPeerError(self, options):
        error = None
        if (typeCheck.isDict(options)):
            error = options.get('hkube_error')

        return error

    def _getFromCacheOrStorage(self, options, dataPath, cacheID):
        data = self._getFromCache(cacheID, dataPath)
        if (data is None):
            size, data = self._getFromStorage(options)
            self._setToCache(cacheID, data, size)
            data = self._getPath(data, dataPath)

        return data

    @trace(name='getFromCache')
    @timing
    def _getFromCache(self, cacheId, dataPath):
        data = self._storageCache.get(cacheId)
        data = self._getPath(data, dataPath)
        return data

    def _setToCache(self, cacheId, data, size):
        self._storageCache.update(cacheId, data, size)

    @trace(name='getFromStorage')
    @timing
    def _getFromStorage(self, options):
        response = self._storageManager.storage.get(options)
        size = len(response)
        return (size, self.decode(response))

    def createStorageInfo(self, options):
        jobId = options.get('jobId')
        taskId = options.get('taskId')
        encodedData = options.get('encodedData')

        path = self._storageManager.hkube.createPath(jobId, taskId)
        metadata = self.createMetadata(options)

        storageInfo = {
            'storageInfo': {
                'path': path,
                'size': len(encodedData) if encodedData else 0
            },
            'metadata': metadata
        }
        return storageInfo

    def createMetadata(self, options):
        nodeName = options.get('nodeName')
        data = options.get('data')
        savePaths = options.get('savePaths', [])

        metadata = dict()
        objData = dict()
        objData[nodeName] = data
        for path in savePaths:
            try:
                value = getPath(objData, path)
                if (value != 'DEFAULT'):
                    meta = self._getMetadata(value)
                    metadata[path] = meta
            except Exception:
                pass

        return metadata

    def _getMetadata(self, value):
        if (typeCheck.isDict(value)):
            meta = {'type': 'object'}
        elif (typeCheck.isList(value)):
            meta = {'type': 'array', 'size': len(value)}
        else:
            meta = {'type': str(type(value).__name__)}
        return meta

    def _getPath(self, data, dataPath):
        if (data and dataPath):
            newData = getPath(data, dataPath)
            if(newData == 'DEFAULT'):
                newData = data
        else:
            newData = data
        return newData





from hkube_python_wrapper.util.object_path import getPath, setPath
import hkube_python_wrapper.util.type_check as typeCheck


def test_getPath_no_path():
    obj = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    path = 'green.no.such'
    result = getPath(obj, path)
    assert result == 'DEFAULT'


def test_getPath_array():
    obj = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    path = '5'
    result = getPath(obj, path)
    assert typeCheck.isInt(result)


def test_getPath_nested_array():
    obj = {"green": {"prop": [1, 2, 3, 4, 5, 6, 7, 8, 9]}}
    path = 'green.prop'
    result = getPath(obj, path)
    assert typeCheck.isList(result)


def test_getPath_nested_array_index():
    obj = {"green": {"prop": [1, 2, 3, 4, 5, 6, 7, 8, 9]}}
    path = 'green.prop.5'
    result = getPath(obj, path)
    assert typeCheck.isInt(result)


def test_getPath_bytes():
    sizeBytes = 10 * 1000000
    obj = {"prop": [{"bytesArr": bytearray(sizeBytes)}]}
    path = 'prop.0.bytesArr'
    result = getPath(obj, path)
    assert result != 'DEFAULT'


def test_setPath_bytes():
    inputArgs = [1, False, {"prop": "$$guid-1234"}]
    value = bytearray(10 * 1000000)
    path = '2.prop'
    setPath(inputArgs, path, value)
    assert typeCheck.isBytearray(inputArgs[2]["prop"])


def test_setPath_array():
    inputArgs = [1, False, {"prop": "$$guid-1234"}]
    value = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    path = '2.prop'
    setPath(inputArgs, path, value)
    assert typeCheck.isList(inputArgs[2]["prop"])



import os


def getIntEnv(name, defaultValue):
    strValue = os.environ.get(name, defaultValue)
    return int(strValue)


socket = {
    "port": os.environ.get('WORKER_SOCKET_PORT', "3001"),
    "host": os.environ.get('WORKER_SOCKET_HOST', "127.0.0.1"),
    "protocol": os.environ.get('WORKER_SOCKET_PROTOCOL', "ws"),
    "url": os.environ.get('WORKER_SOCKET_URL', None),
    "encoding": os.environ.get('WORKER_ALGORITHM_ENCODING', 'json'),
}
discovery = {
    "host": os.environ.get('POD_NAME', '127.0.0.1'),
    "port": os.environ.get('DISCOVERY_PORT', "9020"),
    "encoding": os.environ.get('DISCOVERY_ENCODING', 'bson'),
    "timeout": getIntEnv('DISCOVERY_TIMEOUT', 5000),
    "networkTimeout": getIntEnv('DISCOVERY_NETWORK_TIMEOUT', 1000),
    "maxCacheSize": getIntEnv('DISCOVERY_MAX_CACHE_SIZE_MB', 400)
}
algorithm = {
    "path": os.environ.get('ALGORITHM_PATH', "algorithm_unique_folder"),
    "entryPoint": os.environ.get('ALGORITHM_ENTRY_POINT', "main.py")
}
storage = {
    "clusterName": os.environ.get('CLUSTER_NAME', 'local'),
    "type": os.environ.get('STORAGE_TYPE', 'fs'),
    "mode": os.environ.get('STORAGE_PROTOCOL', 'v2'),
    "encoding": os.environ.get('STORAGE_ENCODING', 'msgpack'),
    "fs": {
        "baseDirectory": os.environ.get('BASE_FS_ADAPTER_DIRECTORY', 'var/tmp/fs/storage')
    },
    "s3": {
        "accessKeyId": os.environ.get('AWS_ACCESS_KEY_ID', 'AKIAIOSFODNN7EXAMPLE'),
        "secretAccessKey": os.environ.get('AWS_SECRET_ACCESS_KEY', 'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'),
        "endpoint": os.environ.get('AWS_ENDPOINT', 'http://127.0.0.1:9000')
    }
}
tracer = {
    "config": {
        'sampler': {
            'type': 'const',
            'param': 1,
        },
        'local_agent': {
            'reporting_host': os.environ.get('JAEGER_AGENT_SERVICE_HOST', 'localhost'),
            'reporting_port': os.environ.get('JAEGER_AGENT_SERVICE_PORT_AGENT_BINARY', os.environ.get('JAEGER_AGENT_SERVICE_PORT', "6831")),
        },
        'logging': True,
    },
    "service_name": os.environ.get('ALGORITHM_NAME', 'algorithm')
}



import sys
if (sys.version_info > (3, 0)):
    # Python 3 code in this block
    from queue import Queue, Empty  # pylint: disable=import-error
else:
    # Python 2 code in this block
    from Queue import Queue, Empty  # pylint: disable=import-error
__all__ = ['Queue', 'Empty']



from __future__ import print_function, division, absolute_import
import hkube_python_wrapper.util.type_check as typeCheck


def getPath(obj, path, defaultValue="DEFAULT"):

    if (not path):
        return obj

    if (obj is None):
        return defaultValue

    if (typeCheck.isString(path)):
        return getPath(obj, path.split('.'), defaultValue)

    currentPath = getKey(path[0])
    nextObj = getShallowProperty(obj, currentPath)
    if (nextObj is None):
        return defaultValue

    if (len(path) == 1):
        return nextObj

    return getPath(obj[currentPath], path[1:], defaultValue)


def getKey(key):
    value = key
    try:
        value = int(key)
    except ValueError:
        pass
    return value


def hasShallowProperty(obj, prop):
    return (typeCheck.isInt(prop) and typeCheck.isList(obj)) or prop in obj


def getShallowProperty(obj, prop):
    if (hasShallowProperty(obj, prop)):
        return obj[prop]
    return None


def setPath(obj, path, value, doNotReplace=False):
    if (typeCheck.isInt(path)):
        path = [path]

    if (not path):
        return obj

    if (typeCheck.isString(path)):
        return setPath(obj, list(map(getKey, path.split('.'))), value, doNotReplace)

    currentPath = path[0]
    currentValue = getShallowProperty(obj, currentPath)
    if (len(path) == 1):
        if (currentValue is None or doNotReplace is False):
            obj[currentPath] = value
        return currentValue

    if (currentValue is None):
        if(typeCheck.isInt(path[1])):
            obj[currentPath] = []
        else:
            obj[currentPath] = {}

    return setPath(obj[currentPath], path[1:], value, doNotReplace)



from threading import Event


class WaitForDataTimeoutException(Exception):
    pass


class WaitForDataStateException(Exception):
    pass


class WaitForData(object):
    def __init__(self, autoreset=False, msg_queue=None):
        self._autoreset = autoreset
        self._msg_queue = msg_queue
        self._event = Event()
        self._event.clear()
        self._data = None

    def reset(self):
        if not self._event.is_set():
            raise WaitForDataStateException()
        self._data = None
        self._event.clear()

    def set(self, data):
        if self._event.is_set():
            raise WaitForDataStateException()
        self._data = data
        self._event.set()

    def get(self, timeout=None):
        if not self._event.wait(timeout):
            raise WaitForDataTimeoutException()
        data = self._data
        if self._autoreset:
            self.reset()
        return data

    def ready(self):
        return self._event.is_set()



import zmq
from .consts import consts
from hkube_python_wrapper.communication.zmq.BaseServer import BaseServer

class ZMQPingServer(BaseServer):
    def __init__(self, context, workerUrl, name):
        self._active = True
        self._socket = None
        self._workerUrl = workerUrl
        self._context = context
        BaseServer.__init__(self, name=name)

    def run(self):
        self._socket = self._context.socket(zmq.REP)
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.connect(self._workerUrl)

        while self._active:
            try:
                events = self._socket.poll(timeout=1000)
                if (events == 0):
                    continue
                message = self._socket.recv()
                if(message == consts.zmq.ping):
                    print('got ping')
                    self._socket.send(consts.zmq.pong)
            except Exception as e:
                print('socket closed: '+str(e))
                break
        print('ZmqPingServer run loop exit')
        self.close()

    def stop(self):
        self._active = False

    def close(self):
        self._socket.close()



from __future__ import print_function, division, absolute_import
import six


def isList(value):
    return isinstance(value, list)


def isString(value):
    return isinstance(value, six.string_types)


def isDict(value):
    return isinstance(value, dict)


def isInt(value):
    return isinstance(value, six.integer_types)


def isBytearray(value):
    return isinstance(value, (bytes, bytearray))



import os
import random
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.storage.storage_manager import StorageManager
from tests.configs import config
from hkube_python_wrapper.util.decorators import timing


def randomString(n):
    min_lc = ord(b'a')
    len_lc = 26
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = min_lc + b % len_lc  # convert 0..255 to 97..122
    return ba.decode("utf-8")


def randomInt(size):
    return random.sample(range(0, size), size)


def createObject(sizeBytes, sizeRandom):
    obj = {
        "bytesData": bytearray(b'\xdd'*(sizeBytes)),
        "anotherBytesData": bytearray(sizeBytes),
        "randomString": randomString(sizeRandom),
        "randomIntArray": randomInt(sizeRandom),
        "dataString": randomString(sizeRandom),
        "bool": False,
        "anotherBool": False,
        "nestedObj": {
            "dataString": randomString(sizeRandom),
            "randomIntArray": randomInt(sizeRandom)
        }
    }
    return obj


encoding = Encoding('msgpack')

storageManager = StorageManager(config.storage)


@timing
def create_bytearray(sizeBytes):
    return bytearray(b'\xdd'*(sizeBytes))


def test_msgpack_bytearray_storage(sizeBytes):
    data = create_bytearray(sizeBytes)

    storageManager.storage.put({"path": "tests/file2", "data": data})
    res = storageManager.storage.get({"path": "tests/file2"})

    data2 = data[0: sizeBytes]
    assert data2 == res

    return data2


def test_msgpack_bytearray(sizeBytes):
    data = create_bytearray(sizeBytes)

    # data = createObject(150, 50)

    encoded = encoding.encode(data)
    leng = len(encoded)

    # assert data == encoded

    decoded = encoding.decode(encoded)

    # data2 = data[0: sizeBytes]
    # assert data2 == decoded

    return encoded


mb = 1024 * 1024

sizes = [10]

for size in sizes:
    test_msgpack_bytearray(size)



from __future__ import print_function, division, absolute_import


class ZmqConsts(object):
    def __init__(self):
        self.ping = b"ping"
        self.pong = b"pong"


class Consts(object):
    def __init__(self):
        self.zmq = ZmqConsts()


consts = Consts()



import zmq
import zmq.devices
import os
import multiprocessing

from .ZMQServer import ZMQServer
from .ZMQPingServer import ZMQPingServer

class ZMQServers(object):
    def __init__(self, port, replyFunc, num_threads):
        self._isServing = False
        self._active = True
        self._replyFunc = replyFunc
        self._url_worker = "inproc://workers"
        self._url_client = "tcp://*:" + str(port)
        self._instances = []
        self._port = int(port)
        self._device = None
        self._context = zmq.Context()
        self._context.setsockopt(zmq.LINGER, 0)
        self._num_threads = num_threads
        self._num_ping_threads = 10

    def listen(self):
        pingProcess = multiprocessing.Process(target=self._createZmqPingServers, args=(self._port,), name="Ping Servers Process")
        pingProcess.daemon = True
        pingProcess.start()
        for _ in range(self._num_threads):
            server = ZMQServer(self._context, self._replyFunc, self._url_worker)
            server.start()
            self._instances.append(server)

        try:
            self._device = zmq.devices.ThreadDevice(zmq.QUEUE, zmq.ROUTER, zmq.DEALER)
            self._device.context_factory = lambda: self._context
            self._device.bind_in(self._url_client)
            self._device.setsockopt_in(zmq.LINGER, 0)
            self._device.bind_out(self._url_worker)
            self._device.setsockopt_out(zmq.LINGER, 0)
            self._device.start()
        except Exception as e:
            print('zmq.device failed with '+str(e))
            raise

    def isServing(self):
        res = any(i.isServing() for i in self._instances)
        return res

    def close(self):
        print('closing zmq servers')

        for i in self._instances:
            print('closing zmq servers - stop')
            i.stop()
        print('joining zmq server threads')
        for i in self._instances:
            i.join()
        print('zmq context closing')
        self._context.term()
        print('zmq context closed')
        print('closed ZmqServers')

    def _createZmqPingServers(self, port):
        try:
            pingContext = zmq.Context()
            url_worker = "inproc://ping_workers"
            url_client = "tcp://*:" + str(port+1)

            def createDealerRouter(url_client, url_worker, context):
                clients = context.socket(zmq.ROUTER)
                clients.bind(url_client)
                workers = context.socket(zmq.DEALER)
                workers.bind(url_worker)
                return (clients, workers)

            clients, workers = createDealerRouter(url_client=url_client, url_worker=url_worker, context=pingContext)
            print('clients: {c}, workers: {w}'.format(c=clients.get(zmq.LAST_ENDPOINT), w=workers.get(zmq.LAST_ENDPOINT)))
            print('Creating {num_threads} ZMQ Ping Servers on port {port}. pid: {pid}'.format(port=url_client, num_threads=self._num_ping_threads, pid=os.getpid()))
            for i in range(self._num_ping_threads):
                server = ZMQPingServer(pingContext, url_worker, 'Ping-Thread-'+str(i))
                server.start()
            zmq.device(zmq.QUEUE, clients, workers)
        except Exception as e:
            print(e)



from __future__ import print_function, division, absolute_import

from ..config import config
from .wc import WebsocketClient
from .data_adapter import DataAdapter
from .methods import methods
from .messages import messages
from hkube_python_wrapper.tracing import Tracer
from hkube_python_wrapper.codeApi.hkube_api import HKubeApi
from hkube_python_wrapper.communication.DataServer import DataServer
from hkube_python_wrapper.util.queueImpl import Queue, Empty
from hkube_python_wrapper.util.timerImpl import Timer
import os
import sys
import importlib
import traceback
from events import Events
from threading import Thread


class Algorunner:
    def __init__(self):
        self._url = None
        self._algorithm = dict()
        self._input = None
        self._events = Events()
        self._loadAlgorithmError = None
        self._connected = False
        self._hkubeApi = None
        self._msg_queue = Queue()
        self._dataAdapter = None
        self._dataServer = None
        self._discovery = None
        self._wsc = None
        self.tracer = None
        self._storage = None
        self._active = True

    @staticmethod
    def Run(start=None, init=None, stop=None, exit=None, options=None):
        """Starts the algorunner wrapper.

        Convenience method to start the algorithm. Pass the algorithm methods
        This method blocks forever

        Args:
            start (function): The entry point of the algorithm. Called for every invocation.
            init (function): Optional init method. Called for every invocation before the start.
            stop (function): Optional stop method. Called when the parent pipeline is stopped.
            exit (function): Optional exit handler. Called before the algorithm is forced to exit.
                    Can be used to clean up resources.

        Returns:
            Never returns.
        """
        algorunner = Algorunner()
        if (start):
            algorunner.loadAlgorithmCallbacks(start, init=init, stop=stop, exit=exit, options=options or config)
        else:
            algorunner.loadAlgorithm(config)
        jobs = algorunner.connectToWorker(config)
        for j in jobs:
            j.join()

    @staticmethod
    def Debug(debug_url, start, init=None, stop=None, exit=None, options=None):
        algorunner = Algorunner()
        config.socket['url'] = debug_url
        config.storage['mode'] = 'v1'
        config.discovery["enable"] = False
        algorunner.loadAlgorithmCallbacks(start, init=init, stop=stop, exit=exit, options=options or config)
        jobs = algorunner.connectToWorker(config)
        for j in jobs:
            j.join()

    def loadAlgorithmCallbacks(self, start, init=None, stop=None, exit=None, options=None):
        try:
            print('Initializing algorithm callbacks')
            self._algorithm['start'] = start
            self._algorithm['init'] = init
            self._algorithm['stop'] = stop
            self._algorithm['exit'] = exit
            for k, v in methods.items():
                methodName = k
                method = v
                isMandatory = method["mandatory"]
                if self._algorithm[methodName] is not None:
                    print('found method {methodName}'.format(methodName=methodName))
                else:
                    mandatory = "mandatory" if isMandatory else "optional"
                    error = 'unable to find {mandatory} method {methodName}'.format(
                        mandatory=mandatory, methodName=methodName)
                    if (isMandatory):
                        raise Exception(error)
                    print(error)
            # fix start if it has only one argument
            if start.__code__.co_argcount == 1:
                self._algorithm['start'] = lambda args, api: start(args)
            self.tracer = Tracer(options.tracer)

        except Exception as e:
            self._loadAlgorithmError = self._errorMsg(e)
            print(e)

    def loadAlgorithm(self, options):
        try:
            cwd = os.getcwd()
            algOptions = options.algorithm
            package = algOptions["path"]
            entry = algOptions["entryPoint"]
            entryPoint = entry.replace("/", ".")
            entryPoint = os.path.splitext(entryPoint)[0]
            __import__(package)
            os.chdir('{cwd}/{package}'.format(cwd=cwd, package=package))
            print('loading {entry}'.format(entry=entry))
            mod = importlib.import_module('.{entryPoint}'.format(
                entryPoint=entryPoint), package=package)
            print('algorithm code loaded')

            for k, v in methods.items():
                methodName = k
                method = v
                isMandatory = method["mandatory"]
                try:
                    self._algorithm[methodName] = getattr(mod, methodName)
                    # fix start if it has only one argument
                    if methodName == 'start' and self._algorithm['start'].__code__.co_argcount == 1:
                        self._algorithm['startOrig'] = self._algorithm['start']
                        self._algorithm['start'] = lambda args, api: self._algorithm['startOrig'](
                            args)
                    print('found method {methodName}'.format(
                        methodName=methodName))
                except Exception as e:
                    mandatory = "mandatory" if isMandatory else "optional"
                    error = 'unable to find {mandatory} method {methodName}'.format(
                        mandatory=mandatory, methodName=methodName)
                    if (isMandatory):
                        raise Exception(error)
                    print(error)
            self.tracer = Tracer(options.tracer)
        except Exception as e:
            self._loadAlgorithmError = self._errorMsg(e)
            traceback.print_exc()
            print(e)

    def connectToWorker(self, options):
        socket = options.socket
        encoding = socket.get("encoding")
        self._storage = options.storage.get("mode")
        url = socket.get("url")

        if (url is not None):
            self._url = url
        else:
            self._url = '{protocol}://{host}:{port}'.format(**socket)

        self._url += '?storage={storage}&encoding={encoding}'.format(
            storage=self._storage, encoding=encoding)

        self._wsc = WebsocketClient(self._msg_queue, encoding, self._url)
        self._initStorage(options)
        self._hkubeApi = HKubeApi(self._wsc, self, self._dataAdapter, self._storage)
        self._registerToWorkerEvents()

        print('connecting to {url}'.format(url=self._url))
        self._wsc.start()
        self._dataServer and self._dataServer.listen()
        runThread = Thread(name="RunThread", target=self.run)
        runThread.start()
        return [self._wsc, runThread]

    def handle(self, command, data):
        if (command == messages.incoming.initialize):
            self._init(data)
        if (command == messages.incoming.start):
            self._start(data)
        if (command == messages.incoming.stop):
            self._stop(data)
        if (command == messages.incoming.exit):
            # call exit on different thread to prevent deadlock
            Timer(0.1, lambda: self._exit(data), name="Exit timer").start()
        if (command in [messages.incoming.algorithmExecutionDone, messages.incoming.algorithmExecutionError]):
            self._hkubeApi.algorithmExecutionDone(data)
        if (command in [messages.incoming.subPipelineDone, messages.incoming.subPipelineError, messages.incoming.subPipelineStopped]):
            self._hkubeApi.subPipelineDone(data)

    def get_message(self, blocking=True):
        return self._msg_queue.get(block=blocking, timeout=0.1)

    def run(self):
        while self._active:
            try:
                (command, data) = self.get_message()
                self.handle(command, data)
            except Empty:
                pass
        print('Exiting run loop')

    def _initStorage(self, options):
        if (self._storage != 'v1'):
            self._initDataServer(options)
        self._initDataAdapter(options)

    def _initDataServer(self, options):
        enable = options.discovery.get("enable")
        if (enable):
            self._discovery = {
                'host': options.discovery.get("host"),
                'port': options.discovery.get("port")
            }
            self._dataServer = DataServer(options.discovery)
            self._reportServing(interval=options.discovery.get("servingReportInterval"))

    def _initDataAdapter(self, options):
        self._dataAdapter = DataAdapter(options, self._dataServer)

    def close(self):
        if (self._wsc):
            self._wsc.shutDown()
        self._active = False

    def _registerToWorkerEvents(self):
        self._wsc.events.on_connection += self._connection
        self._wsc.events.on_disconnect += self._disconnect
        self._wsc.events.on_init += self._init
        self._wsc.events.on_start += self._start
        self._wsc.events.on_stop += self._stop
        self._wsc.events.on_exit += self._exit

    def _connection(self):
        self._connected = True
        print('connected to ' + self._url)

    def _disconnect(self):
        if self._connected:
            print('disconnected from ' + self._url)
        self._connected = False

    def _getMethod(self, name):
        return self._algorithm.get(name)

    def _init(self, options):
        try:
            if (self._loadAlgorithmError):
                self._sendError(self._loadAlgorithmError)
            else:
                self._input = options
                method = self._getMethod('init')
                if (method is not None):
                    method(options)

                self._sendCommand(messages.outgoing.initialized, None)

        except Exception as e:
            self._sendError(e)

    def _start(self, options):
        # pylint: disable=unused-argument
        span = None
        try:
            self._sendCommand(messages.outgoing.started, None)
            # TODO: add parent span from worker
            jobId = self._input.get("jobId")
            taskId = self._input.get("taskId")
            nodeName = self._input.get("nodeName")
            info = self._input.get("info", {})
            savePaths = info.get("savePaths", [])
            topSpan = options.get('spanId') if options else self._input.get('spanId')
            span = Tracer.instance.create_span("start", topSpan, jobId, taskId, nodeName)

            newInput = self._dataAdapter.getData(self._input)
            self._input.update({'input': newInput})
            method = self._getMethod('start')
            algorithmData = method(self._input, self._hkubeApi)
            self._handle_response(algorithmData, jobId, taskId, nodeName, savePaths, span)

        except Exception as e:
            traceback.print_exc()
            Tracer.instance.finish_span(span, e)
            self._sendError(e)

    def _handle_response(self, algorithmData, jobId, taskId, nodeName, savePaths, span):
        if (self._storage == 'v2'):
            self._handle_responseV2(algorithmData, jobId, taskId, nodeName, savePaths, span)
        else:
            self._handle_responseV1(algorithmData, span)

    def _handle_responseV1(self, algorithmData, span):
        if (span):
            Tracer.instance.finish_span(span)
        self._sendCommand(messages.outgoing.done, algorithmData)

    def _handle_responseV2(self, algorithmData, jobId, taskId, nodeName, savePaths, span):
        header, encodedData = self._dataAdapter.encode_separately(algorithmData)
        data = {
            'jobId': jobId,
            'taskId': taskId,
            'nodeName': nodeName,
            'data': algorithmData,
            'encodedData': encodedData,
            'savePaths': savePaths
        }
        storingData = dict()
        storageInfo = self._dataAdapter.createStorageInfo(data)
        storingData.update(storageInfo)
        incache = None
        if (self._dataServer and savePaths):
            incache = self._dataServer.setSendingState(taskId, header, encodedData, len(encodedData))
        encodedData = header + encodedData
        if (incache):
            storingData.update({'discovery': self._discovery, 'taskId': taskId})
            self._sendCommand(messages.outgoing.storing, storingData)
            self._dataAdapter.setData({'jobId': jobId, 'taskId': taskId, 'data': encodedData})
        else:
            self._dataAdapter.setData({'jobId': jobId, 'taskId': taskId, 'data': encodedData})
            self._sendCommand(messages.outgoing.storing, storingData)
        if (span):
            Tracer.instance.finish_span(span)
        self._sendCommand(messages.outgoing.done, None)

    def _reportServing(self, interval=None):
        if (interval is None):
            return
        interval = interval / 1000

        def reportInterval():
            if (not self._active):
                return
            self._reportServingStatus()
            Timer(interval, reportInterval, name='reportIntervalTimer').start()

        reportInterval()

    def _reportServingStatus(self):
        isServing = self._dataServer.isServing()
        if (isServing):
            self._sendCommand(messages.outgoing.servingStatus, True)

    def _stop(self, options):
        try:
            method = self._getMethod('stop')
            if (method is not None):
                method(options)

            self._sendCommand(messages.outgoing.stopped, None)

        except Exception as e:
            self._sendError(e)

    def _exit(self, options):
        try:
            self._dataServer and self._dataServer.shutDown()
            self._wsc.shutDown()
            method = self._getMethod('exit')
            if (method is not None):
                method(options)

            option = options if options is not None else dict()
            code = option.get('exitCode', 0)
            self.close()
            print('Got exit command. Exiting with code', code)
            sys.exit(code)

        except Exception as e:
            print('Got error during exit: '+e)
            # pylint: disable=protected-access
            os._exit(0)

    def _sendCommand(self, command, data):
        try:
            self._wsc.send({'command': command, 'data': data})
        except Exception as e:
            self._sendError(e)

    def _sendError(self, error):
        try:
            print(error)
            self._wsc.send({
                'command': messages.outgoing.error,
                'error': {
                    'code': 'Failed',
                    'message': self._errorMsg(error)
                }
            })
        except Exception as e:
            print(e)

    def _errorMsg(self, error):
        return str(error)




import time
from hkube_python_wrapper.storage.storage_manager import StorageManager
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper import Algorunner
from tests.configs import config
from hkube_python_wrapper.tracing.tracer import Tracer

algorithmName = 'eval-alg'
subpipelineName = 'simple'

encoding = Encoding(config.storage['encoding'])
inp1 = [5, 10]
inp2 = {'d': [6, 'stam']}


def start(args, hkubeApi=None):
    # print('start called')
    waiter1 = hkubeApi.start_algorithm(algorithmName, inp1)
    waiter2 = hkubeApi.start_stored_subpipeline(subpipelineName, inp2)
    res = [waiter1, waiter2]
    return res


def test_callback():
    sm = StorageManager(config.storage)
    algorunner = Algorunner()
    algorunner.loadAlgorithmCallbacks(start, options=config)
    algorunner.connectToWorker(config)
    time.sleep(3)
    res = sm.storage.get({"path": "local-hkube/jobId/taskId"})
    decoded = encoding.decode(res)
    assert decoded[0] == inp1
    assert decoded[1] == inp2
    time.sleep(2)
    Tracer.instance.close()
    algorunner.close()



import datetime
from pympler import asizeof
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.util.decorators import timing


class Cache:
    def __init__(self, config):
        self._cache = dict()
        self._maxCacheSize = config.get('maxCacheSize')
        self.sumSize = 0

    @timing
    def update(self, key, value, size=None, header=None):
        if (size is None):
            if (typeCheck.isBytearray(value)):
                size = len(value)
            else:
                size = asizeof.asizeof(value)
        if (key in self._cache):
            return key
        while (self.sumSize + size) >= self._maxCacheSize * 1000 * 1000:
            if not (self._cache.keys()):
                print("Trying to insert a value of size " + str(size) + " bytes, larger than " + str(
                    self._maxCacheSize) + "MB")
                return None
            self._remove_oldest()
        self._cache[key] = {'timestamp': datetime.datetime.now(), 'size': size, 'value': value, 'header': header}
        self.sumSize += size
        return key

    def __contains__(self, key):
        return key in self._cache

    def _remove_oldest(self):
        oldest = None
        for key in self._cache:
            if oldest is None:
                oldest = key
            elif self._cache[key]['timestamp'] < self._cache[oldest]['timestamp']:
                oldest = key
        self.sumSize -= self._cache[oldest]['size']
        self._cache.pop(oldest)

    def get(self, key):
        item = self._cache.get(key)
        if (item is not None):
            return item.get('value')
        return None

    def getWithHeader(self, key):
        item = self._cache.get(key)
        if (item is not None):
            return item.get('header'), item.get('value')
        return None

    def getAll(self, keys):
        tasksNotInCache = []
        valuesInCache = []
        for key in keys:
            cacheRecord = self._cache.get(key)
            if (cacheRecord):
                valuesInCache.append(cacheRecord.get('value'))
            else:
                tasksNotInCache.append(key)
        return tasksNotInCache, valuesInCache



from __future__ import print_function, division, absolute_import


class Outgoing(object):
    def __init__(self):
        self.initialized = "initialized"
        self.started = "started"
        self.stopped = "stopped"
        self.progress = "progress"
        self.error = "errorMessage"
        self.storing = "storing"
        self.done = "done"
        self.servingStatus = "servingStatus"
        self.startAlgorithmExecution = "startAlgorithmExecution"
        self.stopAlgorithmExecution = "stopAlgorithmExecution"
        self.startRawSubPipeline = "startRawSubPipeline"
        self.startStoredSubPipeline = "startStoredSubPipeline"
        self.stopSubPipeline = "stopSubPipeline"


class Incoming(object):
    def __init__(self):
        self.initialize = "initialize"
        self.start = "start"
        self.stop = "stop"
        self.exit = "exit"
        self.algorithmExecutionError = "algorithmExecutionError"
        self.algorithmExecutionDone = "algorithmExecutionDone"
        self.subPipelineStarted = "subPipelineStarted"
        self.subPipelineError = "subPipelineError"
        self.subPipelineDone = "subPipelineDone"
        self.subPipelineStopped = "subPipelineStopped"


class Messages(object):
    def __init__(self):
        self.outgoing = Outgoing()
        self.incoming = Incoming()


messages = Messages()



from __future__ import print_function, division, absolute_import
import time

import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.wrapper.messages import messages
from .execution import Execution
from .waitFor import WaitForData
from hkube_python_wrapper.util.queueImpl import Empty


class HKubeApi:
    """Hkube interface for code-api operations"""

    def __init__(self, wc, wrapper, dataAdapter, storage):
        self._wc = wc
        self._wrapper = wrapper
        self._dataAdapter = dataAdapter
        self._storage = storage
        self._executions = {}
        self._lastExecId = 0

    def _generateExecId(self):
        self._lastExecId += 1
        return str(self._lastExecId)

    def algorithmExecutionDone(self, data):
        execId = data.get('execId')
        self._handleExecutionDone(execId, data)

    def subPipelineDone(self, data):
        subPipelineId = data.get('subPipelineId')
        self._handleExecutionDone(subPipelineId, data)

    def _handleExecutionDone(self, execId, data):
        execution = self._executions.get(execId)

        try:
            error = data.get('error')
            if(error):
                execution.waiter.set(error)

            elif(execution.includeResult):
                response = data.get('response')
                result = response
                if(typeCheck.isDict(response) and response.get('storageInfo') and self._storage == 'v2'):
                    result = self._dataAdapter.tryGetDataFromPeerOrStorage(response)
                execution.waiter.set(result)
            else:
                execution.waiter.set(None)

        except Exception as e:
            execution.waiter.set(e)
        finally:
            self._executions.pop(execId)

    def start_algorithm(self, algorithmName, input=[], includeResult=True):
        """Starts algorithm execution.

    starts an invocation of algorithm with input, and waits for results

    Args:
        algorithmName (string): The name of the algorithm to start.
        input (array): Optional input for the algorithm.
        includeResult (bool): if True, returns the result of the algorithm execution.
    Returns:
        Returns the result of the algorithm
    """
        print('start_algorithm called with {name}'.format(name=algorithmName))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startAlgorithmExecution,
            "data": {
                "execId": execId,
                "algorithmName": algorithmName,
                "input": input,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)

        return self._waitForResult(execution)

    def start_stored_subpipeline(self, name, flowInput={}, includeResult=True):
        """Starts pipeline execution.

    starts an invocation of a sub-pipeline with input, and waits for results

    Args:
        name (string): The name of the pipeline to start.
        flowInput (dict): Optional flowInput for the pipeline.
        includeResult (bool): if True, returns the result of the pipeline execution.
            default: True
    Returns:
        Returns the result of the pipeline
    """
        print('start_stored_subpipeline called with {name}'.format(name=name))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startStoredSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)
        return self._waitForResult(execution)

    def start_raw_subpipeline(self, name, nodes, flowInput, options={}, webhooks={}, includeResult=True):
        """Starts pipeline execution.

    starts an invocation of a sub-pipeline with input, nodes, options, and optionally waits for results

    Args:
        name (string): The name of the pipeline to start.
        nodes (array): array of node definitions (like in the pipeline descriptor)
        options (dict): pipeline options (like in the pipeline descriptor)
        webhooks (dict): webhook options (like in the pipeline descriptor)
        flowInput (dict): flowInput for the pipeline.
        includeResult (bool): if True, returns the result of the pipeline execution.
            default: True
    Returns:
        Returns the result of the pipeline
    """
        print('start_raw_subpipeline called with {name}'.format(name=name))
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startRawSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "nodes": nodes,
                    "options": options,
                    "webhooks": webhooks,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)
        return self._waitForResult(execution)

    def _waitForResult(self, execution):
        while not execution.waiter.ready():
            try:
                (command, data) = self._wrapper.get_message(False)
                self._wrapper.handle(command, data)
            except Empty:
                pass
            time.sleep(0.01)
        return execution.waiter.get()



import threading
import sys
if (sys.version_info > (3, 0)):
    # Python 3 code in this block
    # pylint: disable=inherit-non-class
    class Timer(threading.Timer):
        def __init__(self, interval, function, args=None, kwargs=None, name='', daemon=True):
            # pylint: disable=non-parent-init-called
            threading.Timer.__init__(self, interval, function, args, kwargs)
            self.name = name
            self.daemon = daemon
else:
    # Python 2 code in this block
    class Timer():
        def __init__(self, interval, function, args=[], kwargs={}, name='', daemon=True):
            self.interval = interval
            self.function = function
            self.args = args
            self.kwargs = kwargs
            self.name = name
            self.daemon = daemon
            self._timer = None

        def start(self):
            self._timer = threading.Timer(self.interval, self.function, self.args, self.kwargs)
            self._timer.name = self.name
            self._timer.daemon = self.daemon
            self._timer.start()




class BaseStorageManager(object):
    def __init__(self, adpter):
        self.adapter = adpter

    def put(self, options):
        try:
            return self.adapter.put(options)
        except Exception:
            raise Exception('Failed to write data to storage')

    def get(self, options):
        try:
            data = self.adapter.get(options)
            if(data is None):
                return None
            return data
        except Exception:
            raise Exception('Failed to read data from storage')


    def list(self, options):
        try:
            return self.adapter.list(options)
        except Exception:
            raise Exception('Failed to list storage data')

    def listPrefix(self, options):
        try:
            return self.adapter.listPrefix(options)
        except Exception:
            raise Exception('Failed to listPrefix storage data')

    def delete(self, options):
        try:
            self.adapter.delete(options)
        except Exception:
            raise Exception('Failed to delete storage data')



import threading
import pytest
from hkube_python_wrapper.communication.DataRequest import DataRequest
from hkube_python_wrapper.communication.DataServer import DataServer
from hkube_python_wrapper.util.encoding import Encoding
from tests.mocks import mockdata
from tests.configs import config

discovery = dict(config.discovery)
discovery.update({"port": 9024})
data1 = mockdata.dataTask1
data2 = mockdata.dataTask2
data2_original = mockdata.dataTask2_original
taskId1 = mockdata.taskId1
taskId2 = mockdata.taskId2
address1 = {'port': "9024", 'host': discovery['host']}

encoding = Encoding(discovery['encoding'])
timeout = 5000
networkTimeout = 1000

ds = DataServer(discovery)
ds.setSendingState(mockdata.taskId1, data1)
ds.setSendingState(mockdata.taskId2, data2)
ds.setSendingState(mockdata.taskId1, mockdata.dataTask1)
ds.setSendingState(mockdata.taskId2, mockdata.dataTask2)

thread = threading.Thread(target=ds.listen)
thread.start()
gevent.sleep(2)


def test_get_data_bytes():
    dr = DataRequest({
        'address': address1,
        'taskId': mockdata.taskId2,
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    print ("reply:"+str(reply))
    assert reply == data2_original


def test_get_data_by_path():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': 'level1',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == data1['level1']
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': 'value1',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == data1['value1']


def test_path_not_exist():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': 'notExist',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == {'hkube_error': {'code': 'noSuchDataPath', 'message': "notExist does not exist in data"}}


def test_get_complete_data():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == data1


def test_data_after_taskid_changed():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == data1
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == data1


def test_success_to_get_data_old_task_id():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == mockdata.dataTask1
    dr = DataRequest({
        'address': address1,
        'taskId': taskId2,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == mockdata.dataTask2_original


def test_failing_no_such_taskid():
    taskId = 'no_such_taskid'
    dr = DataRequest({
        'address': address1,
        'taskId': taskId,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == {'hkube_error': {'code': 'notAvailable', 'message': 'taskId notAvailable'}}


def test_isServing():
    ds = DataServer(discovery)
    gevent.spawn(ds.listen)
    gevent.sleep(0.5)
    assert ds.isServing() == False


def test_shutDown():
    ds = DataServer(discovery)
    gevent.spawn(ds.listen)
    gevent.sleep(0.5)
    ds.shutDown()
    assert ds.isServing() == False


def xtest_waitTillServingEnds():
    def sleepNow(message):
        gevent.sleep(3)
        return ds._createReply(message)

    ds._adapter.getReplyFunc = sleepNow
    dr = DataRequest(
        {'address': address1, 'taskId': taskId1, 'dataPath': 'level1',
         'encoding': 'bson', 'timeout': timeout,
         'networkTimeout': networkTimeout})
    gevent.spawn(dr.invoke)
    gevent.sleep(1)
    assert ds.isServing() == True
    ds.shutDown()
    assert ds.isServing() == False


def test_fail_on_timeout():
    dr = DataRequest({
        'address': address1,
        'taskId': taskId1,
        'dataPath': '',
        'encoding': discovery['encoding'],
        'timeout': timeout,
        'networkTimeout': networkTimeout
    })
    reply = dr.invoke()
    assert reply == {'hkube_error': {'code': 'unknown', 'message': 'Timed out:5000'}}


if __name__ == '__main__':
    test_get_data_bytes()



import os
import pytest
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.storage.storage_manager import StorageManager
from tests.configs import config

config = config.storage.copy()
config['type'] = 's3'

bucket = 'local-hkube'
encoding = Encoding(config['encoding'])

raw = {"data": 'all_my_data'}
encoded = encoding.encode({"data": 'all_my_data'})
sm = StorageManager(config)

@pytest.fixture(scope="session", autouse=True)
def beforeall():
    sm.storage.adapter.init({'bucket': bucket})


def test_put_get():
    options = {'path': bucket + os.path.sep + 'key1', 'data': encoded}
    sm.storage.put(options)
    a = sm.storage.get(options)
    assert encoding.decode(a) == raw


def test_fail_to_get():
    options = {'path': bucket + os.path.sep + 'no_such_key', 'data': encoded}

    with pytest.raises(Exception, match='Failed to read data from storage'):
        sm.storage.get(options)


def test_list():
    bucketList = bucket + '-list'
    sm.storage.adapter.createBucket({'bucket': bucketList})
    options = {'path': bucketList + os.path.sep + 'list_key_1', 'data': encoded}
    sm.storage.put(options)
    options = {'path': bucketList + os.path.sep + 'list_key_2', 'data': encoded}
    sm.storage.put(options)
    options = {'path': bucketList + os.path.sep + 'innerList' + os.path.sep + 'list1', 'data': encoded}
    sm.storage.put(options)
    options = {'path': bucketList}
    resultArr = sm.storage.list(options)
    assert resultArr.__len__() == 3


def test_task_output_put_get():
    obj_path = sm.hkube.put('myJobId', 'myTaksId', encoded)
    assert obj_path == {'path': bucket + '/myJobId/myTaksId'}
    a = sm.hkube.get('myJobId', 'myTaksId')
    assert encoding.decode(a) == raw



import os
import time
# from mock import patch
from time import sleep
from threading import Thread
from hkube_python_wrapper import Algorunner
from tests.configs import config
from tests.mocks import mockdata

oneMB = 1024 * 1024

class Algorithm(object):
    pass


def startCallbackBytes(args):
    return bytearray(b'\xdd'*(1 * oneMB))


def startCallback(args):
    return args["input"]["input"][0]


def test_load_algorithm_callbacks():
    algorunner = Algorunner()
    algorunner.loadAlgorithmCallbacks(startCallback, options=config)
    result1 = algorunner._algorithm['start']({'input': mockdata.initData}, None)
    result2 = startCallback({'input': mockdata.initData})
    assert result1 == result2
    algorunner.close()



def xtest_exit():
    with patch('sys.exit') as exit_mock:
        def doExit(a):
            status['exit'] = True

        def invokeExit():
            algorunner._exit(None)

        def isServingTrue():
            return True

        def isServingFalse():
            return False
        algorunner = Algorunner()
        algorunner.loadAlgorithmCallbacks(startCallback)
        algorunner.connectToWorker(config)
        sleep(1)
        status = {'exit': False}
        algorunner.loadAlgorithmCallbacks(startCallback, exit=doExit)
        algorunner._dataServer.isServing = isServingTrue
        Thread(target=invokeExit).start()
        sleep(1)
        assert status['exit'] == False
        algorunner._dataServer.isServing = isServingFalse
        sleep(1)
        assert status['exit'] == True
        assert exit_mock.called


def test_failed_load_algorithm():
    alg = Algorithm()
    alg.algorithm = {
        "path": "no_such_path",
        "entryPoint": "main.py"
        }
    algorunner = Algorunner()
    algorunner.loadAlgorithm(alg)
    assert "No module named" in algorunner._loadAlgorithmError
    assert "no_such_path" in algorunner._loadAlgorithmError
    algorunner.close()



def xtest_load_algorithm():
    alg = Algorithm()
    alg.algorithm = {
        "path": "test_alg",
        "entryPoint": "main.py"
    }
    cwd = os.getcwd()
    os.chdir(cwd + '/tests')
    algorunner = Algorunner()
    algorunner.loadAlgorithm(alg)

    # os.chdir(cwd)
    result1 = algorunner._algorithm['start']({'input': mockdata.initData}, None)
    result2 = startCallback({'input': mockdata.initData})
    assert result1 == result2

def startCallback2(args):
    return args["input"][0]

def test_connect_to_worker():
    config.discovery.update({"port": "9021"})
    algorunner = Algorunner()
    algorunner.loadAlgorithmCallbacks(startCallback2, options=config)
    algorunner.connectToWorker(config)
    time.sleep(2)
    assert algorunner._connected == True
    assert algorunner._input == mockdata.initData
    algorunner.close()




import time
from hkube_python_wrapper.communication.zmq.BaseServer import BaseServer
import zmq
from .consts import consts


class ZMQServer(BaseServer):
    def __init__(self, context, replyFunc, workerUrl):
        self._lastServing = None
        self._active = True
        self._socket = None
        self._replyFunc = replyFunc
        self._workerUrl = workerUrl
        self._context = context
        BaseServer.__init__(self)

    def run(self):
        self._socket = self._context.socket(zmq.REP)
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.connect(self._workerUrl)

        while self._active:
            try:
                events = self._socket.poll(timeout=1000)
                if (events == 0):
                    continue
                message = self._socket.recv()
                self._lastServing = time.time()
                if(message == consts.zmq.ping):
                    self._socket.send(consts.zmq.pong)
                else:
                    self._send(message)
                self._lastServing = time.time()
            except Exception as e:
                print('socket closed: '+str(e))
                break
        print('ZmqServer run loop exit')
        self.close()

    def _send(self, message):
        try:
            toBeSent = self._replyFunc(message)
            self._socket.send_multipart(toBeSent, copy=False)
        except Exception as e:
            print(e)

    def isServing(self):
        return (self._lastServing) and (time.time() - self._lastServing < 10)

    def stop(self):
        self._active = False

    def close(self):
        self._socket.close()



import os
import boto3


class S3Adapter:
    def __init__(self, config):
        accessKeyId = config["accessKeyId"]
        secretAccessKey = config["secretAccessKey"]
        endpoint = config["endpoint"]

        self.client = boto3.client(
            's3',
            aws_access_key_id=accessKeyId,
            aws_secret_access_key=secretAccessKey,
            endpoint_url=endpoint
        )

    def init(self, options):
        return self.createBucket(options)

    def createBucket(self, options):
        bucket = options["bucket"]
        if(self.isBucketExist(bucket) is False):
            return self.client.create_bucket(Bucket=bucket)
        return None

    def isBucketExist(self, bucket):
        try:
            self.client.head_bucket(Bucket=bucket)
            return True
        except Exception as e:
            if (hasattr(e, 'response') and e.response['Error']['Code'] == '404'):
                return False
            raise e

    def put(self, options):
        path = options["path"]
        body = options["data"]
        parsedPath = self._parsePath(path)
        bucket = parsedPath["bucket"]
        key = parsedPath["key"]
        self.client.put_object(Bucket=bucket, Key=key, Body=body)
        return {'path': bucket + os.path.sep + key}

    def get(self, options):
        path = options["path"]
        parsedPath = self._parsePath(path)
        bucket = parsedPath["bucket"]
        key = parsedPath["key"]
        response = self.client.get_object(Bucket=bucket, Key=key)
        data = response['Body'].read()
        return data

    def list(self, options):
        path = options["path"]
        parsedPath = self._parsePath(path)
        bucket = parsedPath["bucket"]
        response = self.client.list_objects_v2(Bucket=bucket)
        data = list(map(lambda x: {"path": bucket + os.path.sep + x['Key']}, response['Contents']))
        return data

    def _parsePath(self, fullPath):
        seperatedPath = fullPath.split(os.path.sep)
        bucket = seperatedPath[0]
        key = seperatedPath[1:]
        return {"bucket": bucket, "key": os.path.sep.join(key)}



from __future__ import print_function, division, absolute_import
from hkube_python_wrapper.codeApi.hkube_api import HKubeApi
from hkube_python_wrapper.wrapper.algorunner import Algorunner
name = "hkube_python_wrapper"



from __future__ import print_function, division, absolute_import
import time
import datetime
import sys
from hkube_python_wrapper.tracing import Tracer

PY3 = sys.version_info[0] == 3


def trace(name=None):
    def func_wrap(func):
        def wrap(*args, **kwargs):
            if (Tracer.instance is None):
                return func(*args, **kwargs)
            func_name = func.__name__ if PY3 else func.func_name
            operation_name = func_name if name is None else name
            with Tracer.instance.tracer.start_active_span(operation_name=operation_name):
                return func(*args, **kwargs)
        return wrap
    return func_wrap


def timing(func):
    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = func(*args, **kwargs)
        time2 = time.time()
        printTime(func, time1, time2)
        return ret
    return wrap


def printTimePY2(func, time1, time2):
    print('%s function took %0.3f ms' % (func.func_name, (time2 - time1) * 1000.0))


def printTimePY3(func, time1, time2):
    print('{:s} {:s} function took {:.3f} ms'.format(timeFormat(), func.__name__, (time2 - time1) * 1000.0))


def timeFormat():
    now = datetime.datetime.now()
    return now.strftime('%Y-%m-%dT%H:%M:%S')


if PY3:
    printTime = printTimePY3
else:
    printTime = printTimePY2



import os
from hkube_python_wrapper.storage.base_storage_manager import BaseStorageManager


class TaskOutputManager(BaseStorageManager):
    storagePrefix = 'hkube'

    def __init__(self, adapter, config):
        super(TaskOutputManager, self).__init__(adapter)
        self.clusterName = config['clusterName']

    def put(self, jobId, taskId, data):
        return super(TaskOutputManager, self).put({'path': self.createPath(jobId, taskId), 'data': data})

    def get(self, jobId, taskId):
        return super(TaskOutputManager, self).get({'path': self.createPath(jobId, taskId)})

    def list(self, jobId):
        return super(TaskOutputManager, self).list({'path': self.createPath(jobId)})

    def delete(self, jobId, taskId=''):
        return super(TaskOutputManager, self).delete({'path': self.createPath(jobId, taskId)})

    def createPath(self, jobId, taskId=''):
        return self.clusterName + '-' + self.storagePrefix + os.path.sep + jobId + os.path.sep + taskId



#!/usr/bin/env python
from __future__ import print_function, division, absolute_import

import setuptools
import os

here = os.path.abspath(os.path.dirname(__file__))

VERSION = '2.0.22'



packages = setuptools.find_packages()

requires = [
    'Events>=0.3',
    'websocket-client>=0.57.0',
    'simplejson',
    'pymongo>=3.10.1',
    'msgpack',
    'boto3',
    "wsaccel",
    "six",
    "pyzmq",
    "jaeger-client==4",
    "pympler"
]

with open("README.md", "r") as f:
    long_description = f.read()

about = {}
with open(os.path.join(here, 'hkube_python_wrapper', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setuptools.setup(
    name=about['__title__'],
    version=VERSION,
    author=about['__author__'],
    author_email=about['__author_email__'],
    description=about['__description__'],
    license=about['__license__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about['__url__'],
    packages=packages,
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




from time import sleep
from hkube_python_wrapper import Algorunner
from tests.configs import config


def test_load_algorithm_callbacks(mocker):
    interval = 500
    config.discovery["servingReportInterval"] = interval
    config.discovery["enable"] = True
    algorunner = Algorunner()
    spy = mocker.spy(algorunner, '_reportServingStatus')
    algorunner._initDataServer(config)
    sleep(0.8)
    assert spy.call_count == 2
    algorunner.close()
    


from hkube_python_wrapper.communication.zmq.ZMQRequest import ZMQRequest
from hkube_python_wrapper.util.encoding import Encoding


class DataRequest:
    ping_times = []

    def __init__(self, reqDetails):
        encoding = reqDetails.get('encoding')
        address = reqDetails.get('address')
        timeout = reqDetails.get('timeout')
        networkTimeout = reqDetails.get('networkTimeout')
        tasks = reqDetails.get('tasks')
        options = {
            u'tasks': tasks
        }
        self.tasks = tasks
        self.encoding = Encoding(encoding)
        content = self.encoding.encode(options, plainEncode=True)
        self.request = dict()
        self.request.update(address)
        self.request.update({"content": content, "timeout": timeout, "networkTimeout": networkTimeout})

    def invoke(self):
        try:
            print('tcp://' + self.request['host'] + ':' + str(self.request['port']))
            adapter = ZMQRequest(self.request)
            responseFrames = adapter.invokeAdapter()
            results = []
            for i in range(0, int(len(responseFrames)/2)):
                header = responseFrames[i*2]
                content = responseFrames[i*2+1]
                decoded = self.encoding.decode_separately(header, content)
                results.append((len(content), decoded))
            DataRequest.ping_times.append(adapter.pingTime)
            return results
        except Exception as e:
            results = []
            for _ in self.tasks:
                results.append((0, self._createError('unknown', str(e))))
            return results
        finally:
            adapter.close()

    def _createError(self, code, message):
        return {'hkube_error': {'code': code, 'message': message}}



import os
import shutil
import pytest
from hkube_python_wrapper.storage.storage_manager import StorageManager
from hkube_python_wrapper.util.encoding import Encoding
from tests.configs import config

def ensure_dir(dirName):
    if not os.path.exists(dirName):
        os.makedirs(dirName)  

config = config.storage
baseDirectory = config["fs"]["baseDirectory"]
rootDirectory = baseDirectory.split('/')[0]
sm = StorageManager(config)
encoding = Encoding(config['encoding'])

dir1 = 'dir1'
dir2 = 'dir2'

raw = {"data": 'all_my_data'}
encoded = encoding.encode({"data": 'all_my_data'})


@pytest.fixture(scope="session", autouse=True)
def beforeall(request):
    ensure_dir(baseDirectory)

    def afterall():
        shutil.rmtree(rootDirectory, ignore_errors=True)

    request.addfinalizer(afterall)


def test_put_get():
    options = {'path': dir1 + os.path.sep + 'a.txt', 'data': encoded}
    sm.storage.put(options)
    a = sm.storage.get(options)
    assert encoding.decode(a) == raw


def test_fail_to_get():
    options = {'path': dir1 + os.path.sep + 'no_such_path.txt', 'data': encoded}
    with pytest.raises(Exception, match='Failed to read data from storage'):
        sm.storage.get(options)

def test_list():
    options = {'path': dir1 + os.path.sep + 'a.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1 + os.path.sep + 'b.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1 + os.path.sep + 'inner' + os.path.sep + 'c.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1}
    resultArr = sm.storage.list(options)
    assert resultArr.__len__() == 3
    assert {'path': '/' + dir1 + '/a.txt'} in resultArr
    assert {'path': '/' + dir1 + '/b.txt'} in resultArr
    assert {'path': '/' + dir1 + '/inner/c.txt'} in resultArr


def test_prefixlist():
    options = {'path': dir1 + os.path.sep + 'a.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1 + os.path.sep + 'b.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1 + os.path.sep + 'inner' + os.path.sep + 'c.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1}
    resultArr = sm.storage.listPrefix(options)
    assert resultArr.__len__() == 2
    assert '/' + dir1 + '/a.txt' in resultArr
    assert '/' + dir1 + '/b.txt' in resultArr
    assert '/' + dir1 + '/inner/c.txt' not in resultArr


def test_list_noneExsistingPath():
    options = {'path': 'noneExisting'}
    result = sm.storage.list(options)
    assert result == None


def test_delete():
    options = {'path': dir1 + os.path.sep + 'a.txt', 'data': encoded}
    sm.storage.put(options)
    options = {'path': dir1 + os.path.sep + 'b.txt', 'data': encoded}
    sm.storage.put(options)
    sm.storage.delete(options)
    options = {'path': dir1}
    resultArr = sm.storage.listPrefix(options)
    assert resultArr.__len__() == 1
    assert '/' + dir1 + '/a.txt' in resultArr
    assert '/' + dir1 + '/b.txt' not in resultArr


def test_task_output_put_get():
    newConfig = config.copy()
    newConfig.update({"clusterName": dir2})
    sm2 = StorageManager(newConfig)
    obj_path = sm2.hkube.put('myJobId', 'myTaksId', encoded)
    assert obj_path == {'path': dir2 + '-hkube/myJobId/myTaksId'}
    a = sm2.hkube.get('myJobId', 'myTaksId')
    assert encoding.decode(a) == raw



from jaeger_client import Config
import opentracing
# needed for python 3.7+ (workaround
# https://github.com/jaegertracing/jaeger-client-python/blob/master/README.md#initialization--configuration)
from tornado import ioloop  # pylint: disable=unused-import


class Tracer(object):
    instance = None

    def __init__(self, tracer_config):
        if (Tracer.instance is not None):
            Tracer.instance.close()

        config = tracer_config.get('config')
        service_name = tracer_config.get('service_name')
        tracerConfig = Config(
            config=config,
            service_name=service_name,
            validate=True
        )
        self.tracer = tracerConfig.new_tracer()
        Tracer.instance = self

    def close(self):
        if (self.tracer is not None):
            self.tracer.close()

    def extract(self, data):
        return self.tracer.extract(opentracing.Format.TEXT_MAP, data)

    def inject(self, span_context, carrier):
        return self.tracer.inject(span_context, opentracing.Format.TEXT_MAP, carrier)

    def create_span(self, operation_name, parent_span, jobId, taskId, nodeName):
        child_of = Tracer.instance.extract(
            parent_span) if parent_span else None
        span = self.tracer.start_active_span(
            operation_name=operation_name, child_of=child_of)
        span.span.set_tag('jobId', jobId)
        span.span.set_tag('taskId', taskId)
        span.span.set_tag('nodeName', nodeName)
        return span

    def finish_span(self, span, error=None):
        if (span is None):
            return
        if (error is not None):
            span.span.set_tag('error', "true")
            span.span.log_kv({'event': 'error', 'error.object': error})
        span.span.finish()



from __future__ import print_function, division, absolute_import
import sys
import struct
import bson
from bson.codec_options import CodecOptions, TypeRegistry
import simplejson as json
import msgpack
import hkube_python_wrapper.util.type_check as typeCheck


def bson_fallback_encoder(value):
    if typeCheck.isBytearray(value):
        return bson.binary.Binary(value)
    return value


type_registry = TypeRegistry(fallback_encoder=bson_fallback_encoder)
codec_options = CodecOptions(type_registry=type_registry)
PY3 = sys.version_info[0] == 3

'''

- Hkube header format: 8 bytes (64 bit)
- Include 2 bytes for magic number and 2 reserved bytes

+---------------------------------------------------------------+
 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7
+---------------------------------------  ----------------------+
|     VERSION    | HEADER_LENGTH  |  DATA_TYPE  | PROTOCOL_TYPE |
+---------------------------------------------------------------+
|      UNUSED    |     UNUSED     |        MAGIC NUMBER         |
+---------------------------------------------------------------+
|                         Payload Data                          |
+---------------------------------------------------------------+
'''

VERSION = 0x01
HEADER_LENGTH = 0x08
PROTOCOL_TYPE_BSON = 0x01
PROTOCOL_TYPE_JSON = 0x02
PROTOCOL_TYPE_MSGPACK = 0x03
DATA_TYPE_RAW = 0x01
DATA_TYPE_ENCODED = 0x02
UNUSED = 0x00
MAGIC_NUMBER = b'HK'


class Encoding:
    def __init__(self, encoding):
        encoders = {
            "bson": {
                "encode": self._bsonEncode,
                "decode": self._bsonDecode,
                "isBinary": True,
                "protocolType": PROTOCOL_TYPE_BSON
            },
            "json": {
                "encode": self._jsonEncode,
                "decode": self._jsonDecode,
                "isBinary": False,
                "protocolType": PROTOCOL_TYPE_JSON
            },
            "msgpack": {
                "encode": self._msgpackEncode,
                "decode": self._msgpackDecode,
                "isBinary": True,
                "protocolType": PROTOCOL_TYPE_MSGPACK
            }
        }
        encoder = encoders[encoding]
        self.type = encoding
        self._encode = encoder["encode"]
        self._decode = encoder["decode"]
        self.isBinary = encoder["isBinary"]
        self.protocolType = encoder["protocolType"]
        self._fromBytes = self._fromBytesPY3 if PY3 else self._fromBytesPY2
        self._toBytes = self._toBytesPY3 if PY3 else self._toBytesPY2

    def encode_separately(self, value):
        if (typeCheck.isBytearray(value)):
            dataType = DATA_TYPE_RAW
            payload = value
        else:
            dataType = DATA_TYPE_ENCODED
            payload = self._encode(value)

        header = self.createHeader(dataType, self.protocolType)
        return header, payload

    def encode(self, value, plainEncode=False):
        if (not self.isBinary or plainEncode is True):
            return self._encode(value)
        header, payload = self.encode_separately(value)
        header += payload
        return header

    def decode_separately(self, header, value):
        dt = bytes(header[2:3])
        dataType = struct.unpack(">B", dt)[0]
        if (dataType == DATA_TYPE_ENCODED):
            payload = self._decode(value)
        else:
            payload = value
        return payload

    def decode(self, value, plainEncode=False):
        if (not self.isBinary or plainEncode is True):
            return self._decode(value)

        if (not typeCheck.isBytearray(value)):
            return value

        view = self._fromBytes(value)
        totalLength = len(view)
        header = bytes(view[0:HEADER_LENGTH])
        mg = bytes(header[-2:])

        if (mg != MAGIC_NUMBER):
            return self._decode(value)

        ftl = bytes(header[1:2])
        headerLength = struct.unpack(">B", ftl)[0]
        data = view[headerLength: totalLength]
        payload = self.decode_separately(header, data)
        return payload

    def _fromBytesPY2(self, value):
        return value

    def _fromBytesPY3(self, value):
        return memoryview(value)

    def _toBytesPY2(self, value):
        return value

    def _toBytesPY3(self, value):
        return value.tobytes()

    def _bsonEncode(self, value):
        return bson.encode({'data': value}, codec_options=codec_options)

    def _bsonDecode(self, value):
        res = bson.decode(value)
        return res.get("data")

    def _bsonEncode(self, value):
        return bson.encode({'data': value}, codec_options=codec_options)

    def _jsonEncode(self, value):
        return json.dumps(value)

    def _jsonDecode(self, value):
        return json.loads(value)

    def _msgpackEncode(self, value):
        return msgpack.packb(value, use_bin_type=True)

    def _msgpackDecode(self, value):
        return msgpack.unpackb(value, raw=False)

    def createHeader(self, dataType, protocolType):
        header = bytearray()
        header.append(VERSION)
        header.append(HEADER_LENGTH)
        header.append(dataType)
        header.append(protocolType)
        header.append(UNUSED)
        header.append(UNUSED)
        header.extend(MAGIC_NUMBER)
        return header
